#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/resource.h>

#ifdef USE_CPU_COUNTER
/* Enable this when x86_64 cycle-count register is available */
# define TICK_SCALE 1.0
#else
/* Fix for your processor speed: */
# define PROCESSOR_SPEED 1.8e9
# define TICK_SCALE (PROCESSOR_SPEED / 1000.0)
#endif

/* Size of the benchmark: */
#define SIZE 1000
#define ITERS 100000

#ifdef INT
typedef int data_t;
#endif
#ifdef FLOAT
typedef float data_t;
#endif
#ifdef DOUBLE
typedef double data_t;
#endif

#ifdef PLUS
# define IDENT 0
# define OPER +
#endif
#ifdef TIMES
# define IDENT 1
# define OPER *
#endif

typedef struct {
  int len;
  data_t* data;
} vec_rec, * vec_ptr;       /* typedefs struct, pointer to struct */

/* Prevent -O3 from turning the code into a no-op: */
#define NO_INLINE __attribute ((__noinline__))

vec_ptr new_vec(int len) {  /* create vector of specified length */
  vec_ptr result = (vec_ptr) malloc(sizeof(vec_rec));
  if(!result) return NULL;  /* cannot allocate storage */
  result->len = len;
  if(len > 0) {
    data_t* data = (data_t*) calloc(len, sizeof(data_t));
    if(!data) {             /* cannot allocate storage */
      free((void*) result);
      return NULL;
    }
    result->data = data;
  }
  else result->data = NULL;
  return result;
}

/* retrieve vector element and store at dest */
int get_vec_element(vec_ptr v, int index, data_t* dest) {
  if(index < 0 || index >= v->len)   /* bounds checking */
    return 0;
  *dest = v->data[index];
  return 1;
}

int vec_length(vec_ptr v) { return v->len; };

data_t* get_vec_start(vec_ptr v) { return v->data; };

#ifndef USE_CPU_COUNTER
long get_ticks(void)
{
  struct rusage use;
  intptr_t s, u;

  do {
    if (!getrusage(RUSAGE_SELF, &use))
      break;
  } while (errno == EINTR);

  s = use.ru_utime.tv_sec + use.ru_stime.tv_sec;
  u = use.ru_utime.tv_usec + use.ru_stime.tv_usec;

  return s * 1000 + u / 1000;
}
#endif

#ifdef USE_CPU_COUNTER
static inline unsigned long long get_ticks(void)
{
  unsigned int lo, hi;
  // RDTSC copies contents of 64-bit TSC into EDX:EAX
  asm volatile("rdtsc" : "=a" (lo), "=d" (hi));
  return (unsigned long long)hi << 32 | lo;
}
#endif

#ifdef COMBINE1
NO_INLINE void combine1(vec_ptr v, data_t* dest) {
  int i;

  *dest = IDENT;
  for(i = 0; i < vec_length(v); i++) {
    data_t val;
    get_vec_element(v, i, &val);
    *dest = *dest OPER val;
  }
}
# define combine combine1
#endif

#ifdef COMBINE2
/* move call to vec_length out of loop */
NO_INLINE void combine2(vec_ptr v, data_t* dest) {
  int i;
  int length = vec_length(v);

  *dest = IDENT;
  for(i = 0; i < length; i++) {
    data_t val;
    get_vec_element(v, i, &val);
    *dest = *dest OPER val;
  }
}
# define combine combine2
#endif

#ifdef COMBINE3
/* direct access to vector data */
NO_INLINE void combine3(vec_ptr v, data_t* dest) {
  int i;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);

  *dest = IDENT;
  for(i = 0; i < length; i++) 
    *dest = *dest OPER data[i];
}
# define combine combine3
#endif

#ifdef COMBINE4
/* accumulate result in local variable */
NO_INLINE void combine4(vec_ptr v, data_t* dest) {
  int i;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc = IDENT;

  for(i = 0; i < length; i++) 
    acc = acc OPER data[i];
  *dest = acc;
}
#define combine combine4
#endif

#ifdef COMBINE5_2
NO_INLINE void combine5_2(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc = IDENT;

  end = length - 1;
  for (i = 0; i < end; i += 2) {
    acc = acc OPER data[i];
    acc = acc OPER data[i+1];    
  }

  for(; i < length; i++) 
    acc = acc OPER data[i];

  *dest = acc;
}
#define combine combine5_2
#endif

#ifdef COMBINE5_4
NO_INLINE void combine5_4(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc = IDENT;

  end = length - 3;
  for (i = 0; i < end; i += 4) {
    acc = acc OPER data[i];
    acc = acc OPER data[i+1];    
    acc = acc OPER data[i+2];    
    acc = acc OPER data[i+3];    
  }

  for(; i < length; i++) 
    acc = acc OPER data[i];

  *dest = acc;
}
#define combine combine5_4
#endif

#ifdef COMBINE6_2
NO_INLINE void combine6_2(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc0 = IDENT, acc1 = IDENT;

  end = length - 1;
  for (i = 0; i < end; i += 2) {
    acc0 = acc0 OPER data[i];
    acc1 = acc1 OPER data[i+1];
  }

  for(; i < length; i++) 
    acc0 = acc0 OPER data[i];

  *dest = acc0 OPER acc1;
}
#define combine combine6_2
#endif

#ifdef COMBINE6_4
NO_INLINE void combine6_4(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc0 = IDENT, acc1 = IDENT;
  data_t acc2 = IDENT, acc3 = IDENT;

  end = length - 3;
  for (i = 0; i < end; i += 4) {
    acc0 = acc0 OPER data[i];
    acc1 = acc1 OPER data[i+1];
    acc2 = acc2 OPER data[i+2];
    acc3 = acc3 OPER data[i+3];
  }

  for(; i < length; i++) 
    acc0 = acc0 OPER data[i];

  *dest = acc0 OPER acc1 OPER acc2 OPER acc3;
}
#define combine combine6_4
#endif

#ifdef COMBINE6_8
NO_INLINE void combine6_8(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc0 = IDENT, acc1 = IDENT;
  data_t acc2 = IDENT, acc3 = IDENT;
  data_t acc4 = IDENT, acc5 = IDENT;
  data_t acc6 = IDENT, acc7 = IDENT;

  end = length - 7;
  for (i = 0; i < end; i += 8) {
    acc0 = acc0 OPER data[i];
    acc1 = acc1 OPER data[i+1];
    acc2 = acc2 OPER data[i+2];
    acc3 = acc3 OPER data[i+3];
    acc4 = acc4 OPER data[i+4];
    acc5 = acc5 OPER data[i+5];
    acc6 = acc6 OPER data[i+6];
    acc7 = acc7 OPER data[i+7];
  }

  for(; i < length; i++) 
    acc0 = acc0 OPER data[i];

  *dest = (acc0 OPER acc1 OPER acc2 OPER acc3
           OPER acc4 OPER acc5 OPER acc6 OPER acc7);
}
#define combine combine6_8
#endif

#ifdef COMBINE7
NO_INLINE void combine7(vec_ptr v, data_t* dest) {
  int i, end;
  int length = vec_length(v);
  data_t* data = get_vec_start(v);
  data_t acc = IDENT;

  end = length - 1;
  for (i = 0; i < end; i += 2) {
    acc = acc OPER (data[i] OPER data[i+1]);
  }

  for(; i < length; i++) 
    acc = acc OPER data[i];

  *dest = acc;
}
#define combine combine7
#endif

int main()
{
  vec_ptr v;
  data_t r;
  long long pre, post;
  double count;
  int i;

  v = new_vec(SIZE);

  pre = get_ticks();

  for (i = 0; i < ITERS; i++) {
    combine(v, &r);
  }

  post = get_ticks();

  printf("%2.2f", ((post - pre) * TICK_SCALE) / (SIZE*ITERS));
}
